
/* 
 ISD1760 v6 - PASTORCICI ROBERT, YO4HFU, 06.11.2024 
 ATTinyCore 2.0 Board support tested
 Feedthrough Disabled during Beacon mode and play loop
 Feedthrough Disabled/Enabled during start/end beep tone

                ATTINY84 - 8MHz
         ATTinyCore - Clockwise pinout
               ________________
              |      |__|      |
           ---| Vcc        Gnd |---
   PTT OUT ---| 10           0 |--- J1
    PTT IN ---| 9            1 |--- J2
           ---| RST          2 |--- A
         B ---| 8            3 |--- MOSI
     CLOCK ---| 7            4 |--- AUX
        SS ---| 6            5 |--- TONE
              |________________|
*/


#include <EEPROM.h>  // Include EEPROM library

#define DEBUG 0     // Set to 0 to disable debug output & ignore serial
#if DEBUG
#define debugPrint(msg) Serial.println(msg)
#define SERIAL_INIT() Serial.begin(9600)
#else
#define debugPrint(msg) 
#define SERIAL_INIT()
#endif 

const int DATA = 3;           // MOSI
const int CLOCK = 7;          // SCK
const int CS = 6;             // SS (Slave Select / Chip Select)

const int selPin = 8;         // SEL input pin
const int LedRecPin = 2;      // LED REC
const int pttInPin = 9;       // PTT input pin
const int pttOutPin = 10;     // PTT output pin 
const int beepPin = 5;        // Beep output pin (Square Wave)

const int analogPin = A4;     // Analog input for variable beep duration
const int jumper1 = 0;       // Digital input for enabling/disabling beep function
const int jumper2 = 1;       // Digital input for selecting beep frequency

// ISD1760 SPI commands [HEX]
const byte CMD_PU[] = {0x01, 0x00};                  // Power-up
const byte CMD_CONFIG[] = {0x65, 0xB0, 0x05};        // APC - feed-through Enabled 
const byte CMD_CONFIG_FTD[] = {0x65, 0xF0, 0x05};   // APC - feed-through Disabled 
const byte CMD_PLAY[] = {0x40, 0x00};  
const byte CMD_STOP[] = {0x02, 0x00};  
const byte CMD_ERASE[] = {0x43, 0x00}; 
const byte CMD_REC[] = {0x41, 0x00}; 

// State variables
enum State { LOOP_PLAY, READY_REC, BEACON, NONE };
State currentState = NONE;

// Timing variables
unsigned long buttonPressStartTime = 0;
bool buttonPressed = false;
bool lastPttState = HIGH;             // Assuming PTT starts high

unsigned long recStartTime = 0;       // To store REC start time
unsigned long trackLength = 0;        // To store track length
bool isRecording = false;             // To track if recording is active

const unsigned long loopDelay = 5000; // Loop delay in milliseconds (RX time)
unsigned long lastPlayTime = 0;       // Last time playback was initiated
bool isPlaying = false;               // To track if playback is active

unsigned long loopStartTime = 0;      // Time when the loop delay started
bool waitingForLoopDelay = false;     // Flag to check if we're in the delay period

unsigned long beepStartTime = 0;
bool isBeeping = false;
unsigned beepDuration = 300;           // Beep duration, variable controlled by AUX trimmer
int beepFrequency = 800;              // Beep frequency in Hz, variable controlled by Jumper 2
bool BEEP_enabled = true;             // Controlled by Jumper 1

bool beaconActive = false;                       // Indicates if the beacon is currently running
bool beaconToneOn = false;                       // Indicates if the tone is currently on
unsigned long beaconStartTime = 0;               // Start time for the tone duration
unsigned long beaconPauseStartTime = 0;          // Start time for the pause duration

const unsigned long beaconToneDuration = 1000;   // Beacon Tone Duration in ms
const unsigned long beaconPauseDuration = 500;   // Beacon Spacing Duration in ms

bool ptt_play = false;
bool ptt_beep = false;
bool ptt_beacon = false;
bool ptt_key = false;
bool ptt_out_state = false;

void setup() {
    pinMode(selPin, INPUT_PULLUP);     // Use pull-up resistor for SEL
    pinMode(jumper1, INPUT_PULLUP);    // Enable pull-up resistor for Jumper 1
    pinMode(jumper2, INPUT_PULLUP);    // Enable pull-up resistor for Jumper 2
    pinMode(pttInPin, INPUT_PULLUP);   // Use pull-up resistor for PTT
    pinMode(pttOutPin, OUTPUT);        // Set PTT output pin
    pinMode(LedRecPin, OUTPUT);        // Set LED REC output pin
    pinMode(beepPin, OUTPUT);          // Set TONE output pin
    
    digitalWrite(pttOutPin, LOW);      // Initialize PTT output to LOW
    digitalWrite(LedRecPin, LOW);      // Initialize LED REC output to LOW

    pinMode(CS, OUTPUT);
    pinMode(DATA, OUTPUT);
    pinMode(CLOCK, OUTPUT);

    // Set SS 
    digitalWrite(CS, HIGH);
    digitalWrite(CLOCK, HIGH);
   
    SERIAL_INIT();
    debugPrint("ISD1760 Initialized");
 
   delay (10);
   powerUp_ISD();   // Send ISD Power UP via SPI
   delay (50);
   config_ISD();   // Send ISD Configuration via SPI
}

void loop() {
    int buttonState = digitalRead(selPin);
    int pttState = digitalRead(pttInPin);   // Read the state of PTT pin

    // Handle SEL button press
    if (buttonState == LOW && !buttonPressed) {
        buttonPressed = true; // Button is pressed
        buttonPressStartTime = millis(); // Start timing
    } 
    else if (buttonState == HIGH && buttonPressed) {
        buttonPressed = false; // Button is released
        unsigned long pressDuration = millis() - buttonPressStartTime; // Calculate press duration

        // Process SEL button, to avoid fake triggering 
        if (pressDuration >= 150) {   
            if (currentState == NONE) {
                activateState(pressDuration);
            } else {
                deactivateState(); // Only handle state change
            }
        }
    }

    // Check for PTT transitions from HIGH to LOW
    if (lastPttState == HIGH && pttState == LOW) { 
        if (currentState == READY_REC) {
            REC(); // Call REC() if ready_rec is active and PTT goes from 1 to 0
        } else if (currentState == LOOP_PLAY) {
            loopPlayDeactivate(); // Deactivate loop_play if PTT goes from HIGH to LOW
        } else if (currentState == BEACON) {
            beaconDeactivate(); // Deactivate beacon if PTT goes from HIGH to LOW
        }
        ptt_key = true;   //PTT key is pressed 
        ptt_out();   // Update pttOutPin state
    }

    // Check for PTT transitions from LOW to HIGH to deactivate ready_rec and call STOP()
    if (lastPttState == LOW && pttState == HIGH) {
            ptt_key = false;  //PTT key is release 
            ptt_out();   // Update pttOutPin state
        if (currentState == READY_REC) {
            STOP(); // Call STOP() when ready_rec is active and PTT transitions from 0 to 1
            EEPROM_track_save(); // Call EEPROM track save,when REC is completed
            readyRecDeactivate(); // Deactivate ready_rec after STOP is called
        }
    }

    // Update the last PTT state
    lastPttState = pttState;

   // Start playback if LOOP_PLAY is active
    if (currentState == LOOP_PLAY) {
        playTrack();
    }
    
   readBeepSettings();
   updateBEEP();
   updateBeacon(); 
}

// Function to handle state activation based on duration of SEL key
void activateState(unsigned long pressDuration) {
    if (pressDuration >= 6000) {
        readyRecActivate();   // Activate recording if pressed more than 6 seconds
    } else if (pressDuration >= 3000) {
        beaconActivate();    // Activate beacon if pressed between 3 to 6 seconds
    } else if (pressDuration >= 500){  
        loopPlayActivate();  // Activate play loop if pressed between 0.5 to 3 seconds
    }
}

// Function to activate Loop Play
void loopPlayActivate() {
    currentState = LOOP_PLAY;
    debugPrint("Loop Play Activated");
    trackLength = readTrackLengthFromEEPROM();  // Read track length from EEPROM
    debugPrint("Track Length Read: ");
    debugPrint(trackLength);                   // Debugging track length
    PLAY();                                    // Call PLAY() to start playback
    isPlaying = true;                          // Set playback flag
    lastPlayTime = millis();                   // Record the time playback started
}

// Function to play the track based on the saved track length
void playTrack() {
    unsigned long currentTime = millis(); // Get the current time

    if (isPlaying) {
        // Check if enough time has elapsed for playback
        if (currentTime - lastPlayTime >= trackLength) {
            debugPrint("Playback Finished");          // Debugging playback finished
            STOP();                                   // Call STOP() after playback is complete
            lastPlayTime = currentTime;               // Update last playback time
            
            // Start the loop delay if not already waiting
            if (!waitingForLoopDelay) {
                loopStartTime = currentTime;          // Record start time for loop delay
                waitingForLoopDelay = true;           // Enter delay period
                ptt_play = false;
                ptt_out();
                debugPrint("RX TIME START#");
            }
        }

        // Check if we are waiting for the loop delay
        if (waitingForLoopDelay) {
            // Check if the loop delay has passed
            if (currentTime - loopStartTime >= loopDelay) {
                waitingForLoopDelay = false;          // Reset delay flag
                debugPrint("RX TIME END ######");
                PLAY();                              // Restart playback 
                lastPlayTime = millis();             // Reset last playback time for the next loop
            }
        }
    }
}

// PLAY function to control ISD
void PLAY() {
    debugPrint("PLAY SPI");
    sendGroup(CMD_PLAY, sizeof(CMD_PLAY));  
    ptt_play = true;            // Set ptt_play flag
    ptt_out();                  // Update pttOutPin state
}

// Function to deactivate Loop Play
void loopPlayDeactivate() {
    debugPrint("Loop Play Deactivated");
    currentState = NONE;       // Reset state
    isPlaying = false;         // Reset playback flag
    STOP ();
    ptt_play = false;          // Clear ptt_play flag
    ptt_out();                 // Update pttOutPin state
}

// Function to handle state deactivation 
void deactivateState() {
    if (currentState == LOOP_PLAY) {
        loopPlayDeactivate();
    } else if (currentState == READY_REC) {
        readyRecDeactivate(); 
    } else if (currentState == BEACON) {
        beaconDeactivate(); 
    }
}

// Function to activate Ready REC
void readyRecActivate() {
    currentState = READY_REC;
    digitalWrite(LedRecPin, HIGH);  // Activate LED REC pin during READY_REC Activate
    debugPrint("Ready REC Activated");
}

// Function to activate Beacon
void beaconActivate() {
    currentState = BEACON;
    debugPrint("Beacon Activated");
    Beacon();                          // Call Beacon function
    
}

// Function to deactivate Ready REC without calling STOP()
void readyRecDeactivate() {
    digitalWrite(LedRecPin, LOW);  // Turn OFF LED REC pin during READY REC Deactivated
    debugPrint("Ready REC Deactivated");
    currentState = NONE;          // Reset state
    isRecording = false;         // Reset recording flag
}

// Erase function for ISD
void Erase() {
     debugPrint("ERASE SPI");
     sendGroup(CMD_ERASE, sizeof(CMD_ERASE)); 
     delay(150); 
}

// REC function to start recording ISD
void REC() {
    Erase();                  // Call Erase before starting REC
    sendGroup(CMD_REC, sizeof(CMD_REC));
    debugPrint("REC SPI");
    isRecording = true;       // Set recording flag
    recStartTime = millis();  // Record the time REC starts
}

// Function to save recorded track leght to EEPROM
void EEPROM_track_save() {
    if (isRecording) {
        trackLength = millis() - recStartTime; // Calculate the duration of the recording
        debugPrint("Recording Duration: ");
        debugPrint(trackLength); // Debugging output
        isRecording = false; // Reset recording flag
    }
    saveTrackLengthToEEPROM(trackLength); // Save track length to EEPROM after recording
}

// Function to stop recording ISD
void STOP() {
  debugPrint("STOP SPI");
  sendGroup(CMD_STOP, sizeof(CMD_STOP));  
}

// Function to save track length to EEPROM
void saveTrackLengthToEEPROM(unsigned long length) {
    debugPrint("Track Length Saved: ");
    debugPrint(length);
    // EEPROM is limited to 512 bytes, store the length across multiple addresses
    EEPROM.put(0, length);  // Save the track length at address 0
    delay(10);             // Add delay to ensure EEPROM write completes
}

// Function to read the track length from EEPROM
unsigned long readTrackLengthFromEEPROM() {
    unsigned long length;
    EEPROM.get(0, length); // Retrieve the track length from EEPROM
    return length;
}

  void readBeepSettings() {
    // Read and remap analog input A0 to 50-600 ms for beep duration
    int analogValue = analogRead(analogPin);
    beepDuration = map(analogValue, 0, 1023, 50, 600);

    // Check Jumper 1 for enabling/disabling beep function
    BEEP_enabled = digitalRead(jumper1) == HIGH;

    // Check Jumper 2 for frequency selection 800 / 1000Hz
    beepFrequency = digitalRead(jumper2) == HIGH ? 800 : 1000;
}

void ptt_out() {  
    // Detect transition from active (true) to inactive (false) state
    if (ptt_play || ptt_key || ptt_beacon) {
        if (!ptt_out_state) {
            ptt_out_state = true;  // Transition to true
            ptt_out_pin();
        }
    } else {
        if (ptt_out_state) {
            ptt_out_state = false;  // Transition to false
            if (BEEP_enabled) {     // Only trigger beep if BEEP is enabled
                BEEP();
            }
            ptt_out_pin(); 
        }
    }  
}

void ptt_out_pin() // Final PTT, OR gate between PTT OUT and PTT BEEP in order to control PTT OUT PIN
{  
    // Check if READY_REC is active to inhibit PTT output
    if (currentState == READY_REC) {
        digitalWrite(pttOutPin, LOW);  // Inhibit pttOutPin when READY_REC is active
    } else {
        // OR gate logic: Set pttOutPin HIGH if ptt_out_state or ptt_beep is active
        if (ptt_out_state || ptt_beep) {
            digitalWrite(pttOutPin, HIGH);
            debugPrint("PTT ON");
        } else {
            digitalWrite(pttOutPin, LOW);
            debugPrint("PTT OFF");
        }
    }
}

void BEEP() {
    if (!isBeeping) {                  // Start the beep only if it's not already beeping
        ptt_beep = true;
        ptt_out_pin();                 // Update the pin state to stay active during beep
        config_FTD_ISD();              // Disable audio feedthrough before beep tone start
        tone(beepPin, beepFrequency);  // Generate tone on beepPin
        debugPrint("BEEP ON");
        beepStartTime = millis();      // Record the start time
        isBeeping = true;              // Set the beeping flag
        

    }
}

void updateBEEP() {
    if (isBeeping && (millis() - beepStartTime >= beepDuration)) {
        noTone(beepPin);               // Stop the beep
        debugPrint("BEEP OFF");
          ptt_beep = false;            // Reset beep state
          ptt_out_pin();               // Update pin state after beep ends
        isBeeping = false;             // Clear the beeping flag
        config_ISD();                  // Enable audio feedthrough after beep completed

    }
}

void Beacon() {
    if (!beaconActive) {                  // Activate beacon only if it's not already active
        beaconActive = true;              // Set the beacon as active
        beaconToneOn = true;              // Start with tone on
        ptt_beacon = true;
        config_FTD_ISD();                 // Disable audio feedthrough
        ptt_out();         
        beaconStartTime = millis();       // Record the start time of the tone
        tone(beepPin, beepFrequency);     // Start tone
        debugPrint("Beacon ON");
    }
}

void beaconDeactivate() {
    if (beaconActive) {                    // Only deactivate if the beacon is active
      if (!BEEP_enabled) {                 // Send "Feedthrough Enabled" only if  BEEP is disabled, to avoid double sending Enable/Disable
                config_ISD();             // Enable audio feedthrough
      }                    
        noTone(beepPin);                  // Stop any active tone
        beaconActive = false;             // Set beacon as inactive
        beaconToneOn = false;             // Reset tone state
        ptt_beacon = false;
        ptt_out();                   
        debugPrint("Beacon OFF");
        currentState = NONE;              // Reset state
    }
}

void updateBeacon() {
    if (beaconActive) {                                // Only update if the beacon is active
        unsigned long currentTime = millis();
        if (beaconToneOn) {                            // If tone is currently on
            if (currentTime - beaconStartTime >= beaconToneDuration) {
                noTone(beepPin);                       // Stop the tone
                beaconToneOn = false;                  // Set tone to off
                beaconPauseStartTime = currentTime;    // Record the start time of the pause
                debugPrint("Beacon PAUSE");
            }
        } else {                                      // If tone is currently off (in pause)
            if (currentTime - beaconPauseStartTime >= beaconPauseDuration) {
                tone(beepPin, beepFrequency);         // Restart the tone
                beaconToneOn = true;                  // Set tone to on
                beaconStartTime = currentTime;        // Reset the start time of the tone
                debugPrint("Beacon ON");
            }
        }
    }
}

// Emulate SPI (Attiny84 hardware SPI is not available)
void spiTransfer(byte data) {
  for (int i = 0; i < 8; i++) {
    // Send LSB first
    digitalWrite (DATA, !!(data & (1 << i)));
    digitalWrite(CLOCK, HIGH); 
    digitalWrite(CLOCK, LOW); 
  }
}

// Function to send a group of commands 
void sendGroup(const byte* commandGroup, int length) {
  digitalWrite(CS, LOW);
  digitalWrite(CLOCK, LOW);  
  for (int i = 0; i < length; i++) {
    spiTransfer(commandGroup[i]);      
  }
  digitalWrite(DATA, LOW);
  digitalWrite(CS, HIGH);
  digitalWrite(CLOCK, HIGH);                              
}

// Function to power up ISD 
void powerUp_ISD() {
  sendGroup(CMD_PU, sizeof(CMD_PU)); 
}

// Function to configure ISD - feedthrouh enabled
void config_ISD() {
  sendGroup(CMD_CONFIG, sizeof(CMD_CONFIG));  
  debugPrint("Config FT Enabled");
}

// Function to configure ISD - feedthrouh disabled
void config_FTD_ISD() {
  sendGroup(CMD_CONFIG_FTD, sizeof(CMD_CONFIG_FTD)); 
  debugPrint("Config FT Disabled"); 
}
